# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union
from datetime import datetime
from typing_extensions import Literal

import httpx

from ...._types import Body, Omit, Query, Headers, NotGiven, omit, not_given
from ...._utils import maybe_transform, async_maybe_transform
from ...._compat import cached_property
from ...._resource import SyncAPIResource, AsyncAPIResource
from ...._response import (
    BinaryAPIResponse,
    AsyncBinaryAPIResponse,
    StreamedBinaryAPIResponse,
    AsyncStreamedBinaryAPIResponse,
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    to_custom_raw_response_wrapper,
    async_to_streamed_response_wrapper,
    to_custom_streamed_response_wrapper,
    async_to_custom_raw_response_wrapper,
    async_to_custom_streamed_response_wrapper,
)
from ....pagination import SyncCursorPage, AsyncCursorPage
from ...._base_client import AsyncPaginator, make_request_options
from ....types.v1.customers import (
    invoice_list_params,
    invoice_retrieve_params,
    invoice_add_charge_params,
    invoice_list_breakdowns_params,
)
from ....types.v1.customers.invoice import Invoice
from ....types.v1.customers.invoice_retrieve_response import InvoiceRetrieveResponse
from ....types.v1.customers.invoice_add_charge_response import InvoiceAddChargeResponse
from ....types.v1.customers.invoice_list_breakdowns_response import InvoiceListBreakdownsResponse

__all__ = ["InvoicesResource", "AsyncInvoicesResource"]


class InvoicesResource(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> InvoicesResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/Metronome-Industries/metronome-python#accessing-raw-response-data-eg-headers
        """
        return InvoicesResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> InvoicesResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/Metronome-Industries/metronome-python#with_streaming_response
        """
        return InvoicesResourceWithStreamingResponse(self)

    def retrieve(
        self,
        *,
        customer_id: str,
        invoice_id: str,
        skip_zero_qty_line_items: bool | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> InvoiceRetrieveResponse:
        """
        Retrieve detailed information for a specific invoice by its unique identifier.
        This endpoint returns comprehensive invoice data including line items, applied
        credits, totals, and billing period details for both finalized and draft
        invoices.

        ### Use this endpoint to:

        - Display historical invoice details in customer-facing dashboards or billing
          portals.
        - Retrieve current month draft invoices to show customers their month-to-date
          spend.
        - Access finalized invoices for historical billing records and payment
          reconciliation.
        - Validate customer pricing and credit applications for customer support
          queries.

        ### Key response fields:

        Invoice status (DRAFT, FINALIZED, VOID) Billing period start and end dates Total
        amount and amount due after credits Detailed line items broken down by:

        - Customer and contract information
        - Invoice line item type
        - Product/service name and ID
        - Quantity consumed
        - Unit and total price
        - Time period for usage-based charges
        - Applied credits or prepaid commitments

        ### Usage guidelines:

        - Draft invoices update in real-time as usage is reported and may change before
          finalization
        - The response includes both usage-based line items (e.g., API calls, data
          processed) and scheduled charges (e.g., monthly subscriptions, commitment
          fees)
        - Credit and commitment applications are shown as separate line items with
          negative amounts
        - For voided invoices, the response will indicate VOID status but retain all
          original line item details

        Args:
          skip_zero_qty_line_items: If set, all zero quantity line items will be filtered out of the response

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        if not invoice_id:
            raise ValueError(f"Expected a non-empty value for `invoice_id` but received {invoice_id!r}")
        return self._get(
            f"/v1/customers/{customer_id}/invoices/{invoice_id}",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {"skip_zero_qty_line_items": skip_zero_qty_line_items},
                    invoice_retrieve_params.InvoiceRetrieveParams,
                ),
            ),
            cast_to=InvoiceRetrieveResponse,
        )

    def list(
        self,
        *,
        customer_id: str,
        credit_type_id: str | Omit = omit,
        ending_before: Union[str, datetime] | Omit = omit,
        limit: int | Omit = omit,
        next_page: str | Omit = omit,
        skip_zero_qty_line_items: bool | Omit = omit,
        sort: Literal["date_asc", "date_desc"] | Omit = omit,
        starting_on: Union[str, datetime] | Omit = omit,
        status: str | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SyncCursorPage[Invoice]:
        """
        Retrieves a paginated list of invoices for a specific customer, with flexible
        filtering options to narrow results by status, date range, credit type, and
        more. This endpoint provides a comprehensive view of a customer's billing
        history and current charges, supporting both real-time billing dashboards and
        historical reporting needs.

        ### Use this endpoint to:

        - Display historical invoice details in customer-facing dashboards or billing
          portals.
        - Retrieve current month draft invoices to show customers their month-to-date
          spend.
        - Access finalized invoices for historical billing records and payment
          reconciliation.
        - Validate customer pricing and credit applications for customer support
          queries.
        - Generate financial reports by filtering invoices within specific date ranges

        ### Key response fields:

        Array of invoice objects containing:

        - Invoice ID and status (DRAFT, FINALIZED, VOID)
        - Invoice type (USAGE, SCHEDULED)
        - Billing period start and end dates
        - Issue date and due date
        - Total amount, subtotal, and amount due
        - Applied credits summary
        - Contract ID reference
        - External billing provider status (if integrated with Stripe, etc.)
        - Pagination metadata `next_page` cursor

        ### Usage guidelines:

        - The endpoint returns invoice summaries; use the Get Invoice endpoint for
          detailed line items
        - Draft invoices are continuously updated as new usage is reported and will show
          real-time spend
        - Results are ordered by creation date descending by default (newest first)
        - When filtering by date range, the filter applies to the billing period, not
          the issue date
        - For customers with many invoices, implement pagination to ensure all results
          are retrieved External billing provider statuses (like Stripe payment status)
          are included when applicable
        - Voided invoices are included in results by default unless filtered out by
          status

        Args:
          credit_type_id: Only return invoices for the specified credit type

          ending_before: RFC 3339 timestamp (exclusive). Invoices will only be returned for billing
              periods that end before this time.

          limit: Max number of results that should be returned

          next_page: Cursor that indicates where the next page of results should start.

          skip_zero_qty_line_items: If set, all zero quantity line items will be filtered out of the response

          sort: Invoice sort order by issued_at, e.g. date_asc or date_desc. Defaults to
              date_asc.

          starting_on: RFC 3339 timestamp (inclusive). Invoices will only be returned for billing
              periods that start at or after this time.

          status: Invoice status, e.g. DRAFT, FINALIZED, or VOID

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return self._get_api_list(
            f"/v1/customers/{customer_id}/invoices",
            page=SyncCursorPage[Invoice],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "credit_type_id": credit_type_id,
                        "ending_before": ending_before,
                        "limit": limit,
                        "next_page": next_page,
                        "skip_zero_qty_line_items": skip_zero_qty_line_items,
                        "sort": sort,
                        "starting_on": starting_on,
                        "status": status,
                    },
                    invoice_list_params.InvoiceListParams,
                ),
            ),
            model=Invoice,
        )

    def add_charge(
        self,
        *,
        customer_id: str,
        charge_id: str,
        customer_plan_id: str,
        description: str,
        invoice_start_timestamp: Union[str, datetime],
        price: float,
        quantity: float,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> InvoiceAddChargeResponse:
        """Add a one time charge to the specified invoice.

        This is a Plans (deprecated)
        endpoint. New clients should implement using Contracts.

        Args:
          charge_id: The Metronome ID of the charge to add to the invoice. Note that the charge must
              be on a product that is not on the current plan, and the product must have only
              fixed charges.

          customer_plan_id: The Metronome ID of the customer plan to add the charge to.

          invoice_start_timestamp: The start_timestamp of the invoice to add the charge to.

          price: The price of the charge. This price will match the currency on the invoice, e.g.
              USD cents.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return self._post(
            f"/v1/customers/{customer_id}/addCharge",
            body=maybe_transform(
                {
                    "charge_id": charge_id,
                    "customer_plan_id": customer_plan_id,
                    "description": description,
                    "invoice_start_timestamp": invoice_start_timestamp,
                    "price": price,
                    "quantity": quantity,
                },
                invoice_add_charge_params.InvoiceAddChargeParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=InvoiceAddChargeResponse,
        )

    def list_breakdowns(
        self,
        *,
        customer_id: str,
        ending_before: Union[str, datetime],
        starting_on: Union[str, datetime],
        credit_type_id: str | Omit = omit,
        limit: int | Omit = omit,
        next_page: str | Omit = omit,
        skip_zero_qty_line_items: bool | Omit = omit,
        sort: Literal["date_asc", "date_desc"] | Omit = omit,
        status: str | Omit = omit,
        window_size: Literal["HOUR", "DAY"] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SyncCursorPage[InvoiceListBreakdownsResponse]:
        """
        Retrieve granular time-series breakdowns of invoice data at hourly or daily
        intervals. This endpoint transforms standard invoices into detailed timelines,
        enabling you to track usage patterns, identify consumption spikes, and provide
        customers with transparency into their billing details throughout the billing
        period.

        ### Use this endpoint to:

        - Build usage analytics dashboards showing daily or hourly consumption trends
        - Identify peak usage periods for capacity planning and cost optimization
        - Generate detailed billing reports for finance teams and customer success
        - Troubleshoot billing disputes by examining usage patterns at specific times
        - Power real-time cost monitoring and alerting systems

        ### Key response fields:

        An array of BreakdownInvoice objects, each containing:

        - All standard invoice fields (ID, customer, commit, line items, totals, status)
        - Line items with quantities and costs for that specific period
        - `breakdown_start_timestamp`: Start of the specific time window
        - `breakdown_end_timestamp`: End of the specific time window
        - `next_page`: Pagination cursor for large result sets

        ### Usage guidelines:

        - Time granularity: Set `window_size` to hour or day based on your analysis
          needs
        - Response limits: Daily breakdowns return up to 35 days; hourly breakdowns
          return up to 24 hours per request
        - Date filtering: Use `starting_on` and `ending_before` to focus on specific
          periods
        - Performance: For large date ranges, use pagination to retrieve all data
          efficiently
        - Backdated usage: If usage events arrive after invoice finalization, breakdowns
          will reflect the updated usage
        - Zero quantity filtering: Use `skip_zero_qty_line_items=true` to exclude
          periods with no usage

        Args:
          ending_before: RFC 3339 timestamp. Breakdowns will only be returned for time windows that end
              on or before this time.

          starting_on: RFC 3339 timestamp. Breakdowns will only be returned for time windows that start
              on or after this time.

          credit_type_id: Only return invoices for the specified credit type

          limit: Max number of results that should be returned. For daily breakdowns, the
              response can return up to 35 days worth of breakdowns. For hourly breakdowns,
              the response can return up to 24 hours. If there are more results, a cursor to
              the next page is returned.

          next_page: Cursor that indicates where the next page of results should start.

          skip_zero_qty_line_items: If set, all zero quantity line items will be filtered out of the response

          sort: Invoice sort order by issued_at, e.g. date_asc or date_desc. Defaults to
              date_asc.

          status: Invoice status, e.g. DRAFT or FINALIZED

          window_size: The granularity of the breakdowns to return. Defaults to day.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return self._get_api_list(
            f"/v1/customers/{customer_id}/invoices/breakdowns",
            page=SyncCursorPage[InvoiceListBreakdownsResponse],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "ending_before": ending_before,
                        "starting_on": starting_on,
                        "credit_type_id": credit_type_id,
                        "limit": limit,
                        "next_page": next_page,
                        "skip_zero_qty_line_items": skip_zero_qty_line_items,
                        "sort": sort,
                        "status": status,
                        "window_size": window_size,
                    },
                    invoice_list_breakdowns_params.InvoiceListBreakdownsParams,
                ),
            ),
            model=InvoiceListBreakdownsResponse,
        )

    def retrieve_pdf(
        self,
        *,
        customer_id: str,
        invoice_id: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> BinaryAPIResponse:
        """Retrieve a PDF version of a specific invoice by its unique identifier.

        This
        endpoint generates a professionally formatted invoice document suitable for
        sharing with customers, accounting teams, or for record-keeping purposes.

        ### Use this endpoint to:

        - Provide customers with downloadable or emailable copies of their invoices
        - Support accounting and finance teams with official billing documents
        - Maintain accurate records of billing transactions for audits and compliance

        ### Key response details:

        - The response is a binary PDF file representing the full invoice
        - The PDF includes all standard invoice information such as line items, totals,
          billing period, and customer details
        - The document is formatted for clarity and professionalism, suitable for
          official use

        ### Usage guidelines:

        - Ensure the `invoice_id` corresponds to an existing invoice for the specified
          `customer_id`
        - The PDF is generated on-demand; frequent requests for the same invoice may
          impact performance
        - Use appropriate headers to handle the binary response in your application
          (e.g., setting `Content-Type: application/pdf`)

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        if not invoice_id:
            raise ValueError(f"Expected a non-empty value for `invoice_id` but received {invoice_id!r}")
        extra_headers = {"Accept": "application/pdf", **(extra_headers or {})}
        return self._get(
            f"/v1/customers/{customer_id}/invoices/{invoice_id}/pdf",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=BinaryAPIResponse,
        )


class AsyncInvoicesResource(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncInvoicesResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/Metronome-Industries/metronome-python#accessing-raw-response-data-eg-headers
        """
        return AsyncInvoicesResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncInvoicesResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/Metronome-Industries/metronome-python#with_streaming_response
        """
        return AsyncInvoicesResourceWithStreamingResponse(self)

    async def retrieve(
        self,
        *,
        customer_id: str,
        invoice_id: str,
        skip_zero_qty_line_items: bool | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> InvoiceRetrieveResponse:
        """
        Retrieve detailed information for a specific invoice by its unique identifier.
        This endpoint returns comprehensive invoice data including line items, applied
        credits, totals, and billing period details for both finalized and draft
        invoices.

        ### Use this endpoint to:

        - Display historical invoice details in customer-facing dashboards or billing
          portals.
        - Retrieve current month draft invoices to show customers their month-to-date
          spend.
        - Access finalized invoices for historical billing records and payment
          reconciliation.
        - Validate customer pricing and credit applications for customer support
          queries.

        ### Key response fields:

        Invoice status (DRAFT, FINALIZED, VOID) Billing period start and end dates Total
        amount and amount due after credits Detailed line items broken down by:

        - Customer and contract information
        - Invoice line item type
        - Product/service name and ID
        - Quantity consumed
        - Unit and total price
        - Time period for usage-based charges
        - Applied credits or prepaid commitments

        ### Usage guidelines:

        - Draft invoices update in real-time as usage is reported and may change before
          finalization
        - The response includes both usage-based line items (e.g., API calls, data
          processed) and scheduled charges (e.g., monthly subscriptions, commitment
          fees)
        - Credit and commitment applications are shown as separate line items with
          negative amounts
        - For voided invoices, the response will indicate VOID status but retain all
          original line item details

        Args:
          skip_zero_qty_line_items: If set, all zero quantity line items will be filtered out of the response

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        if not invoice_id:
            raise ValueError(f"Expected a non-empty value for `invoice_id` but received {invoice_id!r}")
        return await self._get(
            f"/v1/customers/{customer_id}/invoices/{invoice_id}",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=await async_maybe_transform(
                    {"skip_zero_qty_line_items": skip_zero_qty_line_items},
                    invoice_retrieve_params.InvoiceRetrieveParams,
                ),
            ),
            cast_to=InvoiceRetrieveResponse,
        )

    def list(
        self,
        *,
        customer_id: str,
        credit_type_id: str | Omit = omit,
        ending_before: Union[str, datetime] | Omit = omit,
        limit: int | Omit = omit,
        next_page: str | Omit = omit,
        skip_zero_qty_line_items: bool | Omit = omit,
        sort: Literal["date_asc", "date_desc"] | Omit = omit,
        starting_on: Union[str, datetime] | Omit = omit,
        status: str | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> AsyncPaginator[Invoice, AsyncCursorPage[Invoice]]:
        """
        Retrieves a paginated list of invoices for a specific customer, with flexible
        filtering options to narrow results by status, date range, credit type, and
        more. This endpoint provides a comprehensive view of a customer's billing
        history and current charges, supporting both real-time billing dashboards and
        historical reporting needs.

        ### Use this endpoint to:

        - Display historical invoice details in customer-facing dashboards or billing
          portals.
        - Retrieve current month draft invoices to show customers their month-to-date
          spend.
        - Access finalized invoices for historical billing records and payment
          reconciliation.
        - Validate customer pricing and credit applications for customer support
          queries.
        - Generate financial reports by filtering invoices within specific date ranges

        ### Key response fields:

        Array of invoice objects containing:

        - Invoice ID and status (DRAFT, FINALIZED, VOID)
        - Invoice type (USAGE, SCHEDULED)
        - Billing period start and end dates
        - Issue date and due date
        - Total amount, subtotal, and amount due
        - Applied credits summary
        - Contract ID reference
        - External billing provider status (if integrated with Stripe, etc.)
        - Pagination metadata `next_page` cursor

        ### Usage guidelines:

        - The endpoint returns invoice summaries; use the Get Invoice endpoint for
          detailed line items
        - Draft invoices are continuously updated as new usage is reported and will show
          real-time spend
        - Results are ordered by creation date descending by default (newest first)
        - When filtering by date range, the filter applies to the billing period, not
          the issue date
        - For customers with many invoices, implement pagination to ensure all results
          are retrieved External billing provider statuses (like Stripe payment status)
          are included when applicable
        - Voided invoices are included in results by default unless filtered out by
          status

        Args:
          credit_type_id: Only return invoices for the specified credit type

          ending_before: RFC 3339 timestamp (exclusive). Invoices will only be returned for billing
              periods that end before this time.

          limit: Max number of results that should be returned

          next_page: Cursor that indicates where the next page of results should start.

          skip_zero_qty_line_items: If set, all zero quantity line items will be filtered out of the response

          sort: Invoice sort order by issued_at, e.g. date_asc or date_desc. Defaults to
              date_asc.

          starting_on: RFC 3339 timestamp (inclusive). Invoices will only be returned for billing
              periods that start at or after this time.

          status: Invoice status, e.g. DRAFT, FINALIZED, or VOID

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return self._get_api_list(
            f"/v1/customers/{customer_id}/invoices",
            page=AsyncCursorPage[Invoice],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "credit_type_id": credit_type_id,
                        "ending_before": ending_before,
                        "limit": limit,
                        "next_page": next_page,
                        "skip_zero_qty_line_items": skip_zero_qty_line_items,
                        "sort": sort,
                        "starting_on": starting_on,
                        "status": status,
                    },
                    invoice_list_params.InvoiceListParams,
                ),
            ),
            model=Invoice,
        )

    async def add_charge(
        self,
        *,
        customer_id: str,
        charge_id: str,
        customer_plan_id: str,
        description: str,
        invoice_start_timestamp: Union[str, datetime],
        price: float,
        quantity: float,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> InvoiceAddChargeResponse:
        """Add a one time charge to the specified invoice.

        This is a Plans (deprecated)
        endpoint. New clients should implement using Contracts.

        Args:
          charge_id: The Metronome ID of the charge to add to the invoice. Note that the charge must
              be on a product that is not on the current plan, and the product must have only
              fixed charges.

          customer_plan_id: The Metronome ID of the customer plan to add the charge to.

          invoice_start_timestamp: The start_timestamp of the invoice to add the charge to.

          price: The price of the charge. This price will match the currency on the invoice, e.g.
              USD cents.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return await self._post(
            f"/v1/customers/{customer_id}/addCharge",
            body=await async_maybe_transform(
                {
                    "charge_id": charge_id,
                    "customer_plan_id": customer_plan_id,
                    "description": description,
                    "invoice_start_timestamp": invoice_start_timestamp,
                    "price": price,
                    "quantity": quantity,
                },
                invoice_add_charge_params.InvoiceAddChargeParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=InvoiceAddChargeResponse,
        )

    def list_breakdowns(
        self,
        *,
        customer_id: str,
        ending_before: Union[str, datetime],
        starting_on: Union[str, datetime],
        credit_type_id: str | Omit = omit,
        limit: int | Omit = omit,
        next_page: str | Omit = omit,
        skip_zero_qty_line_items: bool | Omit = omit,
        sort: Literal["date_asc", "date_desc"] | Omit = omit,
        status: str | Omit = omit,
        window_size: Literal["HOUR", "DAY"] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> AsyncPaginator[InvoiceListBreakdownsResponse, AsyncCursorPage[InvoiceListBreakdownsResponse]]:
        """
        Retrieve granular time-series breakdowns of invoice data at hourly or daily
        intervals. This endpoint transforms standard invoices into detailed timelines,
        enabling you to track usage patterns, identify consumption spikes, and provide
        customers with transparency into their billing details throughout the billing
        period.

        ### Use this endpoint to:

        - Build usage analytics dashboards showing daily or hourly consumption trends
        - Identify peak usage periods for capacity planning and cost optimization
        - Generate detailed billing reports for finance teams and customer success
        - Troubleshoot billing disputes by examining usage patterns at specific times
        - Power real-time cost monitoring and alerting systems

        ### Key response fields:

        An array of BreakdownInvoice objects, each containing:

        - All standard invoice fields (ID, customer, commit, line items, totals, status)
        - Line items with quantities and costs for that specific period
        - `breakdown_start_timestamp`: Start of the specific time window
        - `breakdown_end_timestamp`: End of the specific time window
        - `next_page`: Pagination cursor for large result sets

        ### Usage guidelines:

        - Time granularity: Set `window_size` to hour or day based on your analysis
          needs
        - Response limits: Daily breakdowns return up to 35 days; hourly breakdowns
          return up to 24 hours per request
        - Date filtering: Use `starting_on` and `ending_before` to focus on specific
          periods
        - Performance: For large date ranges, use pagination to retrieve all data
          efficiently
        - Backdated usage: If usage events arrive after invoice finalization, breakdowns
          will reflect the updated usage
        - Zero quantity filtering: Use `skip_zero_qty_line_items=true` to exclude
          periods with no usage

        Args:
          ending_before: RFC 3339 timestamp. Breakdowns will only be returned for time windows that end
              on or before this time.

          starting_on: RFC 3339 timestamp. Breakdowns will only be returned for time windows that start
              on or after this time.

          credit_type_id: Only return invoices for the specified credit type

          limit: Max number of results that should be returned. For daily breakdowns, the
              response can return up to 35 days worth of breakdowns. For hourly breakdowns,
              the response can return up to 24 hours. If there are more results, a cursor to
              the next page is returned.

          next_page: Cursor that indicates where the next page of results should start.

          skip_zero_qty_line_items: If set, all zero quantity line items will be filtered out of the response

          sort: Invoice sort order by issued_at, e.g. date_asc or date_desc. Defaults to
              date_asc.

          status: Invoice status, e.g. DRAFT or FINALIZED

          window_size: The granularity of the breakdowns to return. Defaults to day.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return self._get_api_list(
            f"/v1/customers/{customer_id}/invoices/breakdowns",
            page=AsyncCursorPage[InvoiceListBreakdownsResponse],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "ending_before": ending_before,
                        "starting_on": starting_on,
                        "credit_type_id": credit_type_id,
                        "limit": limit,
                        "next_page": next_page,
                        "skip_zero_qty_line_items": skip_zero_qty_line_items,
                        "sort": sort,
                        "status": status,
                        "window_size": window_size,
                    },
                    invoice_list_breakdowns_params.InvoiceListBreakdownsParams,
                ),
            ),
            model=InvoiceListBreakdownsResponse,
        )

    async def retrieve_pdf(
        self,
        *,
        customer_id: str,
        invoice_id: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> AsyncBinaryAPIResponse:
        """Retrieve a PDF version of a specific invoice by its unique identifier.

        This
        endpoint generates a professionally formatted invoice document suitable for
        sharing with customers, accounting teams, or for record-keeping purposes.

        ### Use this endpoint to:

        - Provide customers with downloadable or emailable copies of their invoices
        - Support accounting and finance teams with official billing documents
        - Maintain accurate records of billing transactions for audits and compliance

        ### Key response details:

        - The response is a binary PDF file representing the full invoice
        - The PDF includes all standard invoice information such as line items, totals,
          billing period, and customer details
        - The document is formatted for clarity and professionalism, suitable for
          official use

        ### Usage guidelines:

        - Ensure the `invoice_id` corresponds to an existing invoice for the specified
          `customer_id`
        - The PDF is generated on-demand; frequent requests for the same invoice may
          impact performance
        - Use appropriate headers to handle the binary response in your application
          (e.g., setting `Content-Type: application/pdf`)

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        if not invoice_id:
            raise ValueError(f"Expected a non-empty value for `invoice_id` but received {invoice_id!r}")
        extra_headers = {"Accept": "application/pdf", **(extra_headers or {})}
        return await self._get(
            f"/v1/customers/{customer_id}/invoices/{invoice_id}/pdf",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=AsyncBinaryAPIResponse,
        )


class InvoicesResourceWithRawResponse:
    def __init__(self, invoices: InvoicesResource) -> None:
        self._invoices = invoices

        self.retrieve = to_raw_response_wrapper(
            invoices.retrieve,
        )
        self.list = to_raw_response_wrapper(
            invoices.list,
        )
        self.add_charge = to_raw_response_wrapper(
            invoices.add_charge,
        )
        self.list_breakdowns = to_raw_response_wrapper(
            invoices.list_breakdowns,
        )
        self.retrieve_pdf = to_custom_raw_response_wrapper(
            invoices.retrieve_pdf,
            BinaryAPIResponse,
        )


class AsyncInvoicesResourceWithRawResponse:
    def __init__(self, invoices: AsyncInvoicesResource) -> None:
        self._invoices = invoices

        self.retrieve = async_to_raw_response_wrapper(
            invoices.retrieve,
        )
        self.list = async_to_raw_response_wrapper(
            invoices.list,
        )
        self.add_charge = async_to_raw_response_wrapper(
            invoices.add_charge,
        )
        self.list_breakdowns = async_to_raw_response_wrapper(
            invoices.list_breakdowns,
        )
        self.retrieve_pdf = async_to_custom_raw_response_wrapper(
            invoices.retrieve_pdf,
            AsyncBinaryAPIResponse,
        )


class InvoicesResourceWithStreamingResponse:
    def __init__(self, invoices: InvoicesResource) -> None:
        self._invoices = invoices

        self.retrieve = to_streamed_response_wrapper(
            invoices.retrieve,
        )
        self.list = to_streamed_response_wrapper(
            invoices.list,
        )
        self.add_charge = to_streamed_response_wrapper(
            invoices.add_charge,
        )
        self.list_breakdowns = to_streamed_response_wrapper(
            invoices.list_breakdowns,
        )
        self.retrieve_pdf = to_custom_streamed_response_wrapper(
            invoices.retrieve_pdf,
            StreamedBinaryAPIResponse,
        )


class AsyncInvoicesResourceWithStreamingResponse:
    def __init__(self, invoices: AsyncInvoicesResource) -> None:
        self._invoices = invoices

        self.retrieve = async_to_streamed_response_wrapper(
            invoices.retrieve,
        )
        self.list = async_to_streamed_response_wrapper(
            invoices.list,
        )
        self.add_charge = async_to_streamed_response_wrapper(
            invoices.add_charge,
        )
        self.list_breakdowns = async_to_streamed_response_wrapper(
            invoices.list_breakdowns,
        )
        self.retrieve_pdf = async_to_custom_streamed_response_wrapper(
            invoices.retrieve_pdf,
            AsyncStreamedBinaryAPIResponse,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional
from typing_extensions import Literal

from ...._models import BaseModel

__all__ = ["PlanListPriceAdjustmentsResponse", "Price"]


class Price(BaseModel):
    adjustment_type: Literal["fixed", "quantity", "percentage", "override"]
    """Determines how the value will be applied."""

    quantity: Optional[float] = None

    tier: Optional[float] = None
    """Used in pricing tiers. Indicates at what metric value the price applies."""

    value: Optional[float] = None


class PlanListPriceAdjustmentsResponse(BaseModel):
    charge_id: str

    charge_type: Literal["usage", "fixed", "composite", "minimum", "seat"]

    prices: List[Price]

    start_period: float

    quantity: Optional[float] = None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Iterable
from typing_extensions import Literal, Required, TypedDict

__all__ = ["AlertRetrieveParams", "GroupValue", "SeatFilter"]


class AlertRetrieveParams(TypedDict, total=False):
    alert_id: Required[str]
    """The Metronome ID of the threshold notification"""

    customer_id: Required[str]
    """The Metronome ID of the customer"""

    group_values: Iterable[GroupValue]
    """Only present for `spend_threshold_reached` notifications.

    Retrieve the notification for a specific group key-value pair.
    """

    plans_or_contracts: Literal["PLANS", "CONTRACTS"]
    """
    When parallel threshold notifications are enabled during migration, this flag
    denotes whether to fetch notifications for plans or contracts.
    """

    seat_filter: SeatFilter
    """Only allowed for `low_remaining_seat_balance_reached` notifications.

    This filters alerts by the seat group key-value pair.
    """


class GroupValue(TypedDict, total=False):
    """
    Scopes threshold notification evaluation to a specific presentation group key on individual line items. Only present for spend notifications.
    """

    key: Required[str]

    value: Required[str]


class SeatFilter(TypedDict, total=False):
    """Only allowed for `low_remaining_seat_balance_reached` notifications.

    This filters alerts by the seat group key-value pair.
    """

    seat_group_key: Required[str]
    """The seat group key (e.g., "seat_id", "user_id")"""

    seat_group_value: Required[str]
    """The specific seat identifier to filter by"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union
from datetime import datetime
from typing_extensions import Literal, Annotated, TypedDict

from ..._types import SequenceNotStr
from ..._utils import PropertyInfo

__all__ = ["CreditGrantListEntriesParams"]


class CreditGrantListEntriesParams(TypedDict, total=False):
    next_page: str
    """Cursor that indicates where the next page of results should start."""

    sort: Literal["asc", "desc"]
    """Ledgers sort order by date, asc or desc. Defaults to asc."""

    credit_type_ids: SequenceNotStr[str]
    """A list of Metronome credit type IDs to fetch ledger entries for.

    If absent, ledger entries for all credit types will be returned.
    """

    customer_ids: SequenceNotStr[str]
    """A list of Metronome customer IDs to fetch ledger entries for.

    If absent, ledger entries for all customers will be returned.
    """

    ending_before: Annotated[Union[str, datetime], PropertyInfo(format="iso8601")]
    """
    If supplied, ledger entries will only be returned with an effective_at before
    this time. This timestamp must not be in the future. If no timestamp is
    supplied, all entries up to the start of the customer's next billing period will
    be returned.
    """

    starting_on: Annotated[Union[str, datetime], PropertyInfo(format="iso8601")]
    """
    If supplied, only ledger entries effective at or after this time will be
    returned.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import pytest

from metronome import Metronome, AsyncMetronome
from tests.utils import assert_matches_type
from metronome.types.v1 import (
    Payment,
    PaymentCancelResponse,
    PaymentAttemptResponse,
)
from metronome.pagination import SyncBodyCursorPage, AsyncBodyCursorPage

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestPayments:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @parametrize
    def test_method_list(self, client: Metronome) -> None:
        payment = client.v1.payments.list(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            invoice_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
        )
        assert_matches_type(SyncBodyCursorPage[Payment], payment, path=["response"])

    @parametrize
    def test_method_list_with_all_params(self, client: Metronome) -> None:
        payment = client.v1.payments.list(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            invoice_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
            limit=1,
            next_page="next_page",
            statuses=["pending", "requires_intervention"],
        )
        assert_matches_type(SyncBodyCursorPage[Payment], payment, path=["response"])

    @parametrize
    def test_raw_response_list(self, client: Metronome) -> None:
        response = client.v1.payments.with_raw_response.list(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            invoice_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        payment = response.parse()
        assert_matches_type(SyncBodyCursorPage[Payment], payment, path=["response"])

    @parametrize
    def test_streaming_response_list(self, client: Metronome) -> None:
        with client.v1.payments.with_streaming_response.list(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            invoice_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            payment = response.parse()
            assert_matches_type(SyncBodyCursorPage[Payment], payment, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_method_attempt(self, client: Metronome) -> None:
        payment = client.v1.payments.attempt(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            invoice_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
        )
        assert_matches_type(PaymentAttemptResponse, payment, path=["response"])

    @parametrize
    def test_raw_response_attempt(self, client: Metronome) -> None:
        response = client.v1.payments.with_raw_response.attempt(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            invoice_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        payment = response.parse()
        assert_matches_type(PaymentAttemptResponse, payment, path=["response"])

    @parametrize
    def test_streaming_response_attempt(self, client: Metronome) -> None:
        with client.v1.payments.with_streaming_response.attempt(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            invoice_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            payment = response.parse()
            assert_matches_type(PaymentAttemptResponse, payment, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_method_cancel(self, client: Metronome) -> None:
        payment = client.v1.payments.cancel(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            invoice_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
        )
        assert_matches_type(PaymentCancelResponse, payment, path=["response"])

    @parametrize
    def test_raw_response_cancel(self, client: Metronome) -> None:
        response = client.v1.payments.with_raw_response.cancel(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            invoice_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        payment = response.parse()
        assert_matches_type(PaymentCancelResponse, payment, path=["response"])

    @parametrize
    def test_streaming_response_cancel(self, client: Metronome) -> None:
        with client.v1.payments.with_streaming_response.cancel(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            invoice_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            payment = response.parse()
            assert_matches_type(PaymentCancelResponse, payment, path=["response"])

        assert cast(Any, response.is_closed) is True


class TestAsyncPayments:
    parametrize = pytest.mark.parametrize(
        "async_client", [False, True, {"http_client": "aiohttp"}], indirect=True, ids=["loose", "strict", "aiohttp"]
    )

    @parametrize
    async def test_method_list(self, async_client: AsyncMetronome) -> None:
        payment = await async_client.v1.payments.list(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            invoice_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
        )
        assert_matches_type(AsyncBodyCursorPage[Payment], payment, path=["response"])

    @parametrize
    async def test_method_list_with_all_params(self, async_client: AsyncMetronome) -> None:
        payment = await async_client.v1.payments.list(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            invoice_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
            limit=1,
            next_page="next_page",
            statuses=["pending", "requires_intervention"],
        )
        assert_matches_type(AsyncBodyCursorPage[Payment], payment, path=["response"])

    @parametrize
    async def test_raw_response_list(self, async_client: AsyncMetronome) -> None:
        response = await async_client.v1.payments.with_raw_response.list(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            invoice_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        payment = await response.parse()
        assert_matches_type(AsyncBodyCursorPage[Payment], payment, path=["response"])

    @parametrize
    async def test_streaming_response_list(self, async_client: AsyncMetronome) -> None:
        async with async_client.v1.payments.with_streaming_response.list(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            invoice_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            payment = await response.parse()
            assert_matches_type(AsyncBodyCursorPage[Payment], payment, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_method_attempt(self, async_client: AsyncMetronome) -> None:
        payment = await async_client.v1.payments.attempt(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            invoice_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
        )
        assert_matches_type(PaymentAttemptResponse, payment, path=["response"])

    @parametrize
    async def test_raw_response_attempt(self, async_client: AsyncMetronome) -> None:
        response = await async_client.v1.payments.with_raw_response.attempt(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            invoice_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        payment = await response.parse()
        assert_matches_type(PaymentAttemptResponse, payment, path=["response"])

    @parametrize
    async def test_streaming_response_attempt(self, async_client: AsyncMetronome) -> None:
        async with async_client.v1.payments.with_streaming_response.attempt(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            invoice_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            payment = await response.parse()
            assert_matches_type(PaymentAttemptResponse, payment, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_method_cancel(self, async_client: AsyncMetronome) -> None:
        payment = await async_client.v1.payments.cancel(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            invoice_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
        )
        assert_matches_type(PaymentCancelResponse, payment, path=["response"])

    @parametrize
    async def test_raw_response_cancel(self, async_client: AsyncMetronome) -> None:
        response = await async_client.v1.payments.with_raw_response.cancel(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            invoice_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        payment = await response.parse()
        assert_matches_type(PaymentCancelResponse, payment, path=["response"])

    @parametrize
    async def test_streaming_response_cancel(self, async_client: AsyncMetronome) -> None:
        async with async_client.v1.payments.with_streaming_response.cancel(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            invoice_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            payment = await response.parse()
            assert_matches_type(PaymentCancelResponse, payment, path=["response"])

        assert cast(Any, response.is_closed) is True



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Required, TypedDict

__all__ = ["CustomerListBillableMetricsParams"]


class CustomerListBillableMetricsParams(TypedDict, total=False):
    customer_id: Required[str]

    include_archived: bool
    """If true, the list of returned metrics will include archived metrics"""

    limit: int
    """Max number of results that should be returned"""

    next_page: str
    """Cursor that indicates where the next page of results should start."""

    on_current_plan: bool
    """
    If true, the list of metrics will be filtered to just ones that are on the
    customer's current plan
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from .invoice import Invoice as Invoice
from .customer_alert import CustomerAlert as CustomerAlert
from .plan_add_params import PlanAddParams as PlanAddParams
from .plan_end_params import PlanEndParams as PlanEndParams
from .plan_list_params import PlanListParams as PlanListParams
from .alert_list_params import AlertListParams as AlertListParams
from .plan_add_response import PlanAddResponse as PlanAddResponse
from .plan_end_response import PlanEndResponse as PlanEndResponse
from .alert_reset_params import AlertResetParams as AlertResetParams
from .commit_list_params import CommitListParams as CommitListParams
from .credit_list_params import CreditListParams as CreditListParams
from .plan_list_response import PlanListResponse as PlanListResponse
from .invoice_list_params import InvoiceListParams as InvoiceListParams
from .commit_create_params import CommitCreateParams as CommitCreateParams
from .credit_create_params import CreditCreateParams as CreditCreateParams
from .alert_retrieve_params import AlertRetrieveParams as AlertRetrieveParams
from .commit_create_response import CommitCreateResponse as CommitCreateResponse
from .credit_create_response import CreditCreateResponse as CreditCreateResponse
from .alert_retrieve_response import AlertRetrieveResponse as AlertRetrieveResponse
from .invoice_retrieve_params import InvoiceRetrieveParams as InvoiceRetrieveParams
from .invoice_add_charge_params import InvoiceAddChargeParams as InvoiceAddChargeParams
from .invoice_retrieve_response import InvoiceRetrieveResponse as InvoiceRetrieveResponse
from .invoice_add_charge_response import InvoiceAddChargeResponse as InvoiceAddChargeResponse
from .billing_config_create_params import BillingConfigCreateParams as BillingConfigCreateParams
from .named_schedule_update_params import NamedScheduleUpdateParams as NamedScheduleUpdateParams
from .commit_update_end_date_params import CommitUpdateEndDateParams as CommitUpdateEndDateParams
from .credit_update_end_date_params import CreditUpdateEndDateParams as CreditUpdateEndDateParams
from .invoice_list_breakdowns_params import InvoiceListBreakdownsParams as InvoiceListBreakdownsParams
from .named_schedule_retrieve_params import NamedScheduleRetrieveParams as NamedScheduleRetrieveParams
from .commit_update_end_date_response import CommitUpdateEndDateResponse as CommitUpdateEndDateResponse
from .credit_update_end_date_response import CreditUpdateEndDateResponse as CreditUpdateEndDateResponse
from .billing_config_retrieve_response import BillingConfigRetrieveResponse as BillingConfigRetrieveResponse
from .invoice_list_breakdowns_response import InvoiceListBreakdownsResponse as InvoiceListBreakdownsResponse
from .named_schedule_retrieve_response import NamedScheduleRetrieveResponse as NamedScheduleRetrieveResponse
from .plan_list_price_adjustments_params import PlanListPriceAdjustmentsParams as PlanListPriceAdjustmentsParams
from .plan_list_price_adjustments_response import PlanListPriceAdjustmentsResponse as PlanListPriceAdjustmentsResponse



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, Optional
from datetime import datetime

from ..._models import BaseModel

__all__ = ["UsageListResponse"]


class UsageListResponse(BaseModel):
    billable_metric_id: str

    billable_metric_name: str

    customer_id: str

    end_timestamp: datetime

    start_timestamp: datetime

    value: Optional[float] = None

    groups: Optional[Dict[str, Optional[float]]] = None
    """Values will be either a number or null.

    Null indicates that there were no matches for the group_by value.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Union
from typing_extensions import TypeAlias

from ..shared.commit import Commit
from ..shared.credit import Credit

__all__ = ["ContractListBalancesResponse"]

ContractListBalancesResponse: TypeAlias = Union[Commit, Credit]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, List, Optional
from datetime import datetime
from typing_extensions import Literal

from ..._models import BaseModel

__all__ = ["CustomerRetrieveBillingConfigurationsResponse", "Data"]


class Data(BaseModel):
    id: str
    """
    ID of this configuration; can be provided as the
    billing_provider_configuration_id when creating a contract.
    """

    archived_at: Optional[datetime] = None

    billing_provider: Literal[
        "aws_marketplace",
        "stripe",
        "netsuite",
        "custom",
        "azure_marketplace",
        "quickbooks_online",
        "workday",
        "gcp_marketplace",
        "metronome",
    ]
    """The billing provider set for this configuration."""

    configuration: Dict[str, object]
    """Configuration for the billing provider.

    The structure of this object is specific to the billing provider.
    """

    customer_id: str

    delivery_method: Literal["direct_to_billing_provider", "aws_sqs", "tackle", "aws_sns"]
    """The method to use for delivering invoices to this customer."""

    delivery_method_configuration: Dict[str, object]
    """Configuration for the delivery method.

    The structure of this object is specific to the delivery method.
    """

    delivery_method_id: str
    """ID of the delivery method to use for this customer."""


class CustomerRetrieveBillingConfigurationsResponse(BaseModel):
    data: List[Data]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, List, Optional
from datetime import datetime
from typing_extensions import Literal

from ...._models import BaseModel
from ...shared.rate import Rate
from ...shared.commit_rate import CommitRate

__all__ = ["RateCardRetrieveRateScheduleResponse", "Data"]


class Data(BaseModel):
    entitled: bool

    product_custom_fields: Dict[str, str]
    """Custom fields to be added eg. { "key1": "value1", "key2": "value2" }"""

    product_id: str

    product_name: str

    product_tags: List[str]

    rate: Rate

    starting_at: datetime

    billing_frequency: Optional[Literal["MONTHLY", "QUARTERLY", "ANNUAL", "WEEKLY"]] = None

    commit_rate: Optional[CommitRate] = None
    """A distinct rate on the rate card.

    You can choose to use this rate rather than list rate when consuming a credit or
    commit.
    """

    ending_before: Optional[datetime] = None

    pricing_group_values: Optional[Dict[str, str]] = None


class RateCardRetrieveRateScheduleResponse(BaseModel):
    data: List[Data]

    next_page: Optional[str] = None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Union, Iterable
from datetime import datetime
from typing_extensions import Literal

import httpx

from ...._types import Body, Omit, Query, Headers, NotGiven, SequenceNotStr, omit, not_given
from ...._utils import maybe_transform, async_maybe_transform
from ...._compat import cached_property
from ...._resource import SyncAPIResource, AsyncAPIResource
from ...._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from ....pagination import SyncBodyCursorPage, AsyncBodyCursorPage
from ...._base_client import AsyncPaginator, make_request_options
from ....types.v1.customers import commit_list_params, commit_create_params, commit_update_end_date_params
from ....types.shared.commit import Commit
from ....types.v1.customers.commit_create_response import CommitCreateResponse
from ....types.shared_params.commit_specifier_input import CommitSpecifierInput
from ....types.v1.customers.commit_update_end_date_response import CommitUpdateEndDateResponse

__all__ = ["CommitsResource", "AsyncCommitsResource"]


class CommitsResource(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> CommitsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/Metronome-Industries/metronome-python#accessing-raw-response-data-eg-headers
        """
        return CommitsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> CommitsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/Metronome-Industries/metronome-python#with_streaming_response
        """
        return CommitsResourceWithStreamingResponse(self)

    def create(
        self,
        *,
        access_schedule: commit_create_params.AccessSchedule,
        customer_id: str,
        priority: float,
        product_id: str,
        type: Literal["PREPAID", "POSTPAID"],
        applicable_contract_ids: SequenceNotStr[str] | Omit = omit,
        applicable_product_ids: SequenceNotStr[str] | Omit = omit,
        applicable_product_tags: SequenceNotStr[str] | Omit = omit,
        custom_fields: Dict[str, str] | Omit = omit,
        description: str | Omit = omit,
        invoice_contract_id: str | Omit = omit,
        invoice_schedule: commit_create_params.InvoiceSchedule | Omit = omit,
        name: str | Omit = omit,
        netsuite_sales_order_id: str | Omit = omit,
        rate_type: Literal["COMMIT_RATE", "LIST_RATE"] | Omit = omit,
        salesforce_opportunity_id: str | Omit = omit,
        specifiers: Iterable[CommitSpecifierInput] | Omit = omit,
        uniqueness_key: str | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> CommitCreateResponse:
        """
        Creates customer-level commits that establish spending commitments for customers
        across their Metronome usage. Commits represent contracted spending obligations
        that can be either prepaid (paid upfront) or postpaid (billed later).

        Note: In most cases, you should add commitments directly to customer contracts
        using the contract/create or contract/edit APIs.

        ### Use this endpoint to:

        Use this endpoint when you need to establish customer-level spending commitments
        that can be applied across multiple contracts or scoped to specific contracts.
        Customer-level commits are ideal for:

        - Enterprise-wide minimum spending agreements that span multiple contracts
        - Multi-contract volume commitments with shared spending pools
        - Cross-contract discount tiers based on aggregate usage

        #### Commit type Requirements:

        - You must specify either "prepaid" or "postpaid" as the commit type:
        - Prepaid commits: Customer pays upfront; invoice_schedule is optional (if
          omitted, creates a commit without an invoice)
        - Postpaid commits: Customer pays when the commitment expires (the end of the
          access_schedule); invoice_schedule is required and must match access_schedule
          totals.

        #### Billing configuration:

        - invoice_contract_id is required for postpaid commits and for prepaid commits
          with billing (only optional for free prepaid commits) unless do_not_invoice is
          set to true
        - For postpaid commits: access_schedule and invoice_schedule must have matching
          amounts
        - For postpaid commits: only one schedule item is allowed in both schedules.

        #### Scoping flexibility:

        Customer-level commits can be configured in a few ways:

        - Contract-specific: Use the `applicable_contract_ids` field to limit the commit
          to specific contracts
        - Cross-contract: Leave `applicable_contract_ids` empty to allow the commit to
          be used across all of the customer's contracts

        #### Product targeting:

        Commits can be scoped to specific products using applicable_product_ids,
        applicable_product_tags, or specifiers, or left unrestricted to apply to all
        products.

        #### Priority considerations:

        When multiple commits are applicable, the one with the lower priority value will
        be consumed first. If there is a tie, contract level commits and credits will be
        applied before customer level commits and credits. Plan your priority scheme
        carefully to ensure commits are applied in the desired order.

        ### Usage guidelines:

        ⚠️ Preferred Alternative: In most cases, you should add commits directly to
        contracts using the create contract or edit contract APIs instead of creating
        customer-level commits. Contract-level commits provide better organization and
        are the recommended approach for standard use cases.

        Args:
          access_schedule: Schedule for distributing the commit to the customer. For "POSTPAID" commits
              only one schedule item is allowed and amount must match invoice_schedule total.

          priority: If multiple credits or commits are applicable, the one with the lower priority
              will apply first.

          product_id: ID of the fixed product associated with the commit. This is required because
              products are used to invoice the commit amount.

          applicable_contract_ids: Which contract the commit applies to. If not provided, the commit applies to all
              contracts.

          applicable_product_ids: Which products the commit applies to. If applicable_product_ids,
              applicable_product_tags or specifiers are not provided, the commit applies to
              all products.

          applicable_product_tags: Which tags the commit applies to. If applicable_product_ids,
              applicable_product_tags or specifiers are not provided, the commit applies to
              all products.

          custom_fields: Custom fields to be added eg. { "key1": "value1", "key2": "value2" }

          description: Used only in UI/API. It is not exposed to end customers.

          invoice_contract_id: The contract that this commit will be billed on. This is required for "POSTPAID"
              commits and for "PREPAID" commits unless there is no invoice schedule above
              (i.e., the commit is 'free'), or if do_not_invoice is set to true.

          invoice_schedule: Required for "POSTPAID" commits: the true up invoice will be generated at this
              time and only one schedule item is allowed; the total must match
              accesss_schedule amount. Optional for "PREPAID" commits: if not provided, this
              will be a "complimentary" commit with no invoice.

          name: displayed on invoices

          netsuite_sales_order_id: This field's availability is dependent on your client's configuration.

          salesforce_opportunity_id: This field's availability is dependent on your client's configuration.

          specifiers: List of filters that determine what kind of customer usage draws down a commit
              or credit. A customer's usage needs to meet the condition of at least one of the
              specifiers to contribute to a commit's or credit's drawdown. This field cannot
              be used together with `applicable_product_ids` or `applicable_product_tags`.

          uniqueness_key: Prevents the creation of duplicates. If a request to create a commit or credit
              is made with a uniqueness key that was previously used to create a commit or
              credit, a new record will not be created and the request will fail with a 409
              error.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._post(
            "/v1/contracts/customerCommits/create",
            body=maybe_transform(
                {
                    "access_schedule": access_schedule,
                    "customer_id": customer_id,
                    "priority": priority,
                    "product_id": product_id,
                    "type": type,
                    "applicable_contract_ids": applicable_contract_ids,
                    "applicable_product_ids": applicable_product_ids,
                    "applicable_product_tags": applicable_product_tags,
                    "custom_fields": custom_fields,
                    "description": description,
                    "invoice_contract_id": invoice_contract_id,
                    "invoice_schedule": invoice_schedule,
                    "name": name,
                    "netsuite_sales_order_id": netsuite_sales_order_id,
                    "rate_type": rate_type,
                    "salesforce_opportunity_id": salesforce_opportunity_id,
                    "specifiers": specifiers,
                    "uniqueness_key": uniqueness_key,
                },
                commit_create_params.CommitCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=CommitCreateResponse,
        )

    def list(
        self,
        *,
        customer_id: str,
        commit_id: str | Omit = omit,
        covering_date: Union[str, datetime] | Omit = omit,
        effective_before: Union[str, datetime] | Omit = omit,
        include_archived: bool | Omit = omit,
        include_balance: bool | Omit = omit,
        include_contract_commits: bool | Omit = omit,
        include_ledgers: bool | Omit = omit,
        limit: int | Omit = omit,
        next_page: str | Omit = omit,
        starting_at: Union[str, datetime] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SyncBodyCursorPage[Commit]:
        """
        Retrieve all commit agreements for a customer, including both prepaid and
        postpaid commitments. This endpoint provides comprehensive visibility into
        contractual spending obligations, enabling you to track commitment utilization
        and manage customer contracts effectively.

        ### Use this endpoint to:

        - Display commitment balances and utilization in customer dashboards
        - Track prepaid commitment drawdown and remaining balances
        - Monitor postpaid commitment progress toward minimum thresholds
        - Build commitment tracking and forecasting tools
        - Show commitment history with optional ledger details
        - Manage rollover balances between contract periods

        ### Key response fields:

        An array of Commit objects containing:

        - Commit type: PREPAID (pay upfront) or POSTPAID (pay at true-up)
        - Rate type: COMMIT_RATE (discounted) or LIST_RATE (standard pricing)
        - Access schedule: When commitment funds become available
        - Invoice schedule: When the customer is billed
        - Product targeting: Which product(s) usage is eligible to draw from this commit
        - Optional ledger entries: Transaction history (if `include_ledgers=true`)
        - Balance information: Current available amount (if `include_balance=true`)
        - Rollover settings: Fraction of unused amount that carries forward

        ### Usage guidelines:

        - Pagination: Results limited to 25 commits per page; use 'next_page' for more
        - Date filtering options:
          - `covering_date`: Commits active on a specific date
          - `starting_at`: Commits with access on/after a date
          - `effective_before`: Commits with access before a date (exclusive)
        - Scope options:
          - `include_contract_commits`: Include contract-level commits (not just
            customer-level)
          - `include_archived`: Include archived commits and commits from archived
            contracts
        - Performance considerations:
          - include_ledgers: Adds detailed transaction history (slower)
          - include_balance: Adds current balance calculation (slower)
        - Optional filtering: Use commit_id to retrieve a specific commit

        Args:
          covering_date: Include only commits that have access schedules that "cover" the provided date

          effective_before: Include only commits that have any access before the provided date (exclusive)

          include_archived: Include archived commits and commits from archived contracts.

          include_balance: Include the balance in the response. Setting this flag may cause the query to be
              slower.

          include_contract_commits: Include commits on the contract level.

          include_ledgers: Include commit ledgers in the response. Setting this flag may cause the query to
              be slower.

          limit: The maximum number of commits to return. Defaults to 25.

          next_page: The next page token from a previous response.

          starting_at: Include only commits that have any access on or after the provided date

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get_api_list(
            "/v1/contracts/customerCommits/list",
            page=SyncBodyCursorPage[Commit],
            body=maybe_transform(
                {
                    "customer_id": customer_id,
                    "commit_id": commit_id,
                    "covering_date": covering_date,
                    "effective_before": effective_before,
                    "include_archived": include_archived,
                    "include_balance": include_balance,
                    "include_contract_commits": include_contract_commits,
                    "include_ledgers": include_ledgers,
                    "limit": limit,
                    "next_page": next_page,
                    "starting_at": starting_at,
                },
                commit_list_params.CommitListParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            model=Commit,
            method="post",
        )

    def update_end_date(
        self,
        *,
        commit_id: str,
        customer_id: str,
        access_ending_before: Union[str, datetime] | Omit = omit,
        invoices_ending_before: Union[str, datetime] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> CommitUpdateEndDateResponse:
        """
        Shortens the end date of a prepaid commit to terminate it earlier than
        originally scheduled. Use this endpoint when you need to cancel or reduce the
        duration of an existing prepaid commit. Only works with prepaid commit types and
        can only move the end date forward (earlier), not extend it.

        ### Usage guidelines:

        To extend commit end dates or make other comprehensive edits, use the 'edit
        commit' endpoint instead.

        Args:
          commit_id: ID of the commit to update. Only supports "PREPAID" commits.

          customer_id: ID of the customer whose commit is to be updated

          access_ending_before: RFC 3339 timestamp indicating when access to the commit will end and it will no
              longer be possible to draw it down (exclusive). If not provided, the access will
              not be updated.

          invoices_ending_before: RFC 3339 timestamp indicating when the commit will stop being invoiced
              (exclusive). If not provided, the invoice schedule will not be updated.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._post(
            "/v1/contracts/customerCommits/updateEndDate",
            body=maybe_transform(
                {
                    "commit_id": commit_id,
                    "customer_id": customer_id,
                    "access_ending_before": access_ending_before,
                    "invoices_ending_before": invoices_ending_before,
                },
                commit_update_end_date_params.CommitUpdateEndDateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=CommitUpdateEndDateResponse,
        )


class AsyncCommitsResource(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncCommitsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/Metronome-Industries/metronome-python#accessing-raw-response-data-eg-headers
        """
        return AsyncCommitsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncCommitsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/Metronome-Industries/metronome-python#with_streaming_response
        """
        return AsyncCommitsResourceWithStreamingResponse(self)

    async def create(
        self,
        *,
        access_schedule: commit_create_params.AccessSchedule,
        customer_id: str,
        priority: float,
        product_id: str,
        type: Literal["PREPAID", "POSTPAID"],
        applicable_contract_ids: SequenceNotStr[str] | Omit = omit,
        applicable_product_ids: SequenceNotStr[str] | Omit = omit,
        applicable_product_tags: SequenceNotStr[str] | Omit = omit,
        custom_fields: Dict[str, str] | Omit = omit,
        description: str | Omit = omit,
        invoice_contract_id: str | Omit = omit,
        invoice_schedule: commit_create_params.InvoiceSchedule | Omit = omit,
        name: str | Omit = omit,
        netsuite_sales_order_id: str | Omit = omit,
        rate_type: Literal["COMMIT_RATE", "LIST_RATE"] | Omit = omit,
        salesforce_opportunity_id: str | Omit = omit,
        specifiers: Iterable[CommitSpecifierInput] | Omit = omit,
        uniqueness_key: str | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> CommitCreateResponse:
        """
        Creates customer-level commits that establish spending commitments for customers
        across their Metronome usage. Commits represent contracted spending obligations
        that can be either prepaid (paid upfront) or postpaid (billed later).

        Note: In most cases, you should add commitments directly to customer contracts
        using the contract/create or contract/edit APIs.

        ### Use this endpoint to:

        Use this endpoint when you need to establish customer-level spending commitments
        that can be applied across multiple contracts or scoped to specific contracts.
        Customer-level commits are ideal for:

        - Enterprise-wide minimum spending agreements that span multiple contracts
        - Multi-contract volume commitments with shared spending pools
        - Cross-contract discount tiers based on aggregate usage

        #### Commit type Requirements:

        - You must specify either "prepaid" or "postpaid" as the commit type:
        - Prepaid commits: Customer pays upfront; invoice_schedule is optional (if
          omitted, creates a commit without an invoice)
        - Postpaid commits: Customer pays when the commitment expires (the end of the
          access_schedule); invoice_schedule is required and must match access_schedule
          totals.

        #### Billing configuration:

        - invoice_contract_id is required for postpaid commits and for prepaid commits
          with billing (only optional for free prepaid commits) unless do_not_invoice is
          set to true
        - For postpaid commits: access_schedule and invoice_schedule must have matching
          amounts
        - For postpaid commits: only one schedule item is allowed in both schedules.

        #### Scoping flexibility:

        Customer-level commits can be configured in a few ways:

        - Contract-specific: Use the `applicable_contract_ids` field to limit the commit
          to specific contracts
        - Cross-contract: Leave `applicable_contract_ids` empty to allow the commit to
          be used across all of the customer's contracts

        #### Product targeting:

        Commits can be scoped to specific products using applicable_product_ids,
        applicable_product_tags, or specifiers, or left unrestricted to apply to all
        products.

        #### Priority considerations:

        When multiple commits are applicable, the one with the lower priority value will
        be consumed first. If there is a tie, contract level commits and credits will be
        applied before customer level commits and credits. Plan your priority scheme
        carefully to ensure commits are applied in the desired order.

        ### Usage guidelines:

        ⚠️ Preferred Alternative: In most cases, you should add commits directly to
        contracts using the create contract or edit contract APIs instead of creating
        customer-level commits. Contract-level commits provide better organization and
        are the recommended approach for standard use cases.

        Args:
          access_schedule: Schedule for distributing the commit to the customer. For "POSTPAID" commits
              only one schedule item is allowed and amount must match invoice_schedule total.

          priority: If multiple credits or commits are applicable, the one with the lower priority
              will apply first.

          product_id: ID of the fixed product associated with the commit. This is required because
              products are used to invoice the commit amount.

          applicable_contract_ids: Which contract the commit applies to. If not provided, the commit applies to all
              contracts.

          applicable_product_ids: Which products the commit applies to. If applicable_product_ids,
              applicable_product_tags or specifiers are not provided, the commit applies to
              all products.

          applicable_product_tags: Which tags the commit applies to. If applicable_product_ids,
              applicable_product_tags or specifiers are not provided, the commit applies to
              all products.

          custom_fields: Custom fields to be added eg. { "key1": "value1", "key2": "value2" }

          description: Used only in UI/API. It is not exposed to end customers.

          invoice_contract_id: The contract that this commit will be billed on. This is required for "POSTPAID"
              commits and for "PREPAID" commits unless there is no invoice schedule above
              (i.e., the commit is 'free'), or if do_not_invoice is set to true.

          invoice_schedule: Required for "POSTPAID" commits: the true up invoice will be generated at this
              time and only one schedule item is allowed; the total must match
              accesss_schedule amount. Optional for "PREPAID" commits: if not provided, this
              will be a "complimentary" commit with no invoice.

          name: displayed on invoices

          netsuite_sales_order_id: This field's availability is dependent on your client's configuration.

          salesforce_opportunity_id: This field's availability is dependent on your client's configuration.

          specifiers: List of filters that determine what kind of customer usage draws down a commit
              or credit. A customer's usage needs to meet the condition of at least one of the
              specifiers to contribute to a commit's or credit's drawdown. This field cannot
              be used together with `applicable_product_ids` or `applicable_product_tags`.

          uniqueness_key: Prevents the creation of duplicates. If a request to create a commit or credit
              is made with a uniqueness key that was previously used to create a commit or
              credit, a new record will not be created and the request will fail with a 409
              error.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._post(
            "/v1/contracts/customerCommits/create",
            body=await async_maybe_transform(
                {
                    "access_schedule": access_schedule,
                    "customer_id": customer_id,
                    "priority": priority,
                    "product_id": product_id,
                    "type": type,
                    "applicable_contract_ids": applicable_contract_ids,
                    "applicable_product_ids": applicable_product_ids,
                    "applicable_product_tags": applicable_product_tags,
                    "custom_fields": custom_fields,
                    "description": description,
                    "invoice_contract_id": invoice_contract_id,
                    "invoice_schedule": invoice_schedule,
                    "name": name,
                    "netsuite_sales_order_id": netsuite_sales_order_id,
                    "rate_type": rate_type,
                    "salesforce_opportunity_id": salesforce_opportunity_id,
                    "specifiers": specifiers,
                    "uniqueness_key": uniqueness_key,
                },
                commit_create_params.CommitCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=CommitCreateResponse,
        )

    def list(
        self,
        *,
        customer_id: str,
        commit_id: str | Omit = omit,
        covering_date: Union[str, datetime] | Omit = omit,
        effective_before: Union[str, datetime] | Omit = omit,
        include_archived: bool | Omit = omit,
        include_balance: bool | Omit = omit,
        include_contract_commits: bool | Omit = omit,
        include_ledgers: bool | Omit = omit,
        limit: int | Omit = omit,
        next_page: str | Omit = omit,
        starting_at: Union[str, datetime] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> AsyncPaginator[Commit, AsyncBodyCursorPage[Commit]]:
        """
        Retrieve all commit agreements for a customer, including both prepaid and
        postpaid commitments. This endpoint provides comprehensive visibility into
        contractual spending obligations, enabling you to track commitment utilization
        and manage customer contracts effectively.

        ### Use this endpoint to:

        - Display commitment balances and utilization in customer dashboards
        - Track prepaid commitment drawdown and remaining balances
        - Monitor postpaid commitment progress toward minimum thresholds
        - Build commitment tracking and forecasting tools
        - Show commitment history with optional ledger details
        - Manage rollover balances between contract periods

        ### Key response fields:

        An array of Commit objects containing:

        - Commit type: PREPAID (pay upfront) or POSTPAID (pay at true-up)
        - Rate type: COMMIT_RATE (discounted) or LIST_RATE (standard pricing)
        - Access schedule: When commitment funds become available
        - Invoice schedule: When the customer is billed
        - Product targeting: Which product(s) usage is eligible to draw from this commit
        - Optional ledger entries: Transaction history (if `include_ledgers=true`)
        - Balance information: Current available amount (if `include_balance=true`)
        - Rollover settings: Fraction of unused amount that carries forward

        ### Usage guidelines:

        - Pagination: Results limited to 25 commits per page; use 'next_page' for more
        - Date filtering options:
          - `covering_date`: Commits active on a specific date
          - `starting_at`: Commits with access on/after a date
          - `effective_before`: Commits with access before a date (exclusive)
        - Scope options:
          - `include_contract_commits`: Include contract-level commits (not just
            customer-level)
          - `include_archived`: Include archived commits and commits from archived
            contracts
        - Performance considerations:
          - include_ledgers: Adds detailed transaction history (slower)
          - include_balance: Adds current balance calculation (slower)
        - Optional filtering: Use commit_id to retrieve a specific commit

        Args:
          covering_date: Include only commits that have access schedules that "cover" the provided date

          effective_before: Include only commits that have any access before the provided date (exclusive)

          include_archived: Include archived commits and commits from archived contracts.

          include_balance: Include the balance in the response. Setting this flag may cause the query to be
              slower.

          include_contract_commits: Include commits on the contract level.

          include_ledgers: Include commit ledgers in the response. Setting this flag may cause the query to
              be slower.

          limit: The maximum number of commits to return. Defaults to 25.

          next_page: The next page token from a previous response.

          starting_at: Include only commits that have any access on or after the provided date

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get_api_list(
            "/v1/contracts/customerCommits/list",
            page=AsyncBodyCursorPage[Commit],
            body=maybe_transform(
                {
                    "customer_id": customer_id,
                    "commit_id": commit_id,
                    "covering_date": covering_date,
                    "effective_before": effective_before,
                    "include_archived": include_archived,
                    "include_balance": include_balance,
                    "include_contract_commits": include_contract_commits,
                    "include_ledgers": include_ledgers,
                    "limit": limit,
                    "next_page": next_page,
                    "starting_at": starting_at,
                },
                commit_list_params.CommitListParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            model=Commit,
            method="post",
        )

    async def update_end_date(
        self,
        *,
        commit_id: str,
        customer_id: str,
        access_ending_before: Union[str, datetime] | Omit = omit,
        invoices_ending_before: Union[str, datetime] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> CommitUpdateEndDateResponse:
        """
        Shortens the end date of a prepaid commit to terminate it earlier than
        originally scheduled. Use this endpoint when you need to cancel or reduce the
        duration of an existing prepaid commit. Only works with prepaid commit types and
        can only move the end date forward (earlier), not extend it.

        ### Usage guidelines:

        To extend commit end dates or make other comprehensive edits, use the 'edit
        commit' endpoint instead.

        Args:
          commit_id: ID of the commit to update. Only supports "PREPAID" commits.

          customer_id: ID of the customer whose commit is to be updated

          access_ending_before: RFC 3339 timestamp indicating when access to the commit will end and it will no
              longer be possible to draw it down (exclusive). If not provided, the access will
              not be updated.

          invoices_ending_before: RFC 3339 timestamp indicating when the commit will stop being invoiced
              (exclusive). If not provided, the invoice schedule will not be updated.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._post(
            "/v1/contracts/customerCommits/updateEndDate",
            body=await async_maybe_transform(
                {
                    "commit_id": commit_id,
                    "customer_id": customer_id,
                    "access_ending_before": access_ending_before,
                    "invoices_ending_before": invoices_ending_before,
                },
                commit_update_end_date_params.CommitUpdateEndDateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=CommitUpdateEndDateResponse,
        )


class CommitsResourceWithRawResponse:
    def __init__(self, commits: CommitsResource) -> None:
        self._commits = commits

        self.create = to_raw_response_wrapper(
            commits.create,
        )
        self.list = to_raw_response_wrapper(
            commits.list,
        )
        self.update_end_date = to_raw_response_wrapper(
            commits.update_end_date,
        )


class AsyncCommitsResourceWithRawResponse:
    def __init__(self, commits: AsyncCommitsResource) -> None:
        self._commits = commits

        self.create = async_to_raw_response_wrapper(
            commits.create,
        )
        self.list = async_to_raw_response_wrapper(
            commits.list,
        )
        self.update_end_date = async_to_raw_response_wrapper(
            commits.update_end_date,
        )


class CommitsResourceWithStreamingResponse:
    def __init__(self, commits: CommitsResource) -> None:
        self._commits = commits

        self.create = to_streamed_response_wrapper(
            commits.create,
        )
        self.list = to_streamed_response_wrapper(
            commits.list,
        )
        self.update_end_date = to_streamed_response_wrapper(
            commits.update_end_date,
        )


class AsyncCommitsResourceWithStreamingResponse:
    def __init__(self, commits: AsyncCommitsResource) -> None:
        self._commits = commits

        self.create = async_to_streamed_response_wrapper(
            commits.create,
        )
        self.list = async_to_streamed_response_wrapper(
            commits.list,
        )
        self.update_end_date = async_to_streamed_response_wrapper(
            commits.update_end_date,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union, Iterable
from datetime import datetime
from typing_extensions import Required, Annotated, TypedDict

from ...._utils import PropertyInfo

__all__ = ["RateCardUpdateParams", "Alias"]


class RateCardUpdateParams(TypedDict, total=False):
    rate_card_id: Required[str]
    """ID of the rate card to update"""

    aliases: Iterable[Alias]
    """Reference this alias when creating a contract.

    If the same alias is assigned to multiple rate cards, it will reference the rate
    card to which it was most recently assigned. It is not exposed to end customers.
    """

    description: str

    name: str
    """Used only in UI/API. It is not exposed to end customers."""


class Alias(TypedDict, total=False):
    name: Required[str]

    ending_before: Annotated[Union[str, datetime], PropertyInfo(format="iso8601")]

    starting_at: Annotated[Union[str, datetime], PropertyInfo(format="iso8601")]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Union, Optional
from typing_extensions import Literal, TypeAlias

from ..._models import BaseModel

__all__ = [
    "HierarchyConfiguration",
    "ParentHierarchyConfiguration",
    "ParentHierarchyConfigurationChild",
    "ParentHierarchyConfigurationParentBehavior",
    "ChildHierarchyConfiguration",
    "ChildHierarchyConfigurationParent",
]


class ParentHierarchyConfigurationChild(BaseModel):
    contract_id: str

    customer_id: str


class ParentHierarchyConfigurationParentBehavior(BaseModel):
    invoice_consolidation_type: Optional[Literal["CONCATENATE", "NONE"]] = None
    """
    Account hierarchy M3 - Indicates the desired behavior of consolidated invoices
    generated by the parent in a customer hierarchy

    **CONCATENATE**: Statements on the invoices of child customers will be appended
    to the consolidated invoice

    **NONE**: Do not generate consolidated invoices
    """


class ParentHierarchyConfiguration(BaseModel):
    children: List[ParentHierarchyConfigurationChild]
    """List of contracts that belong to this parent."""

    parent_behavior: Optional[ParentHierarchyConfigurationParentBehavior] = None


class ChildHierarchyConfigurationParent(BaseModel):
    """The single parent contract/customer for this child."""

    contract_id: str

    customer_id: str


class ChildHierarchyConfiguration(BaseModel):
    parent: ChildHierarchyConfigurationParent
    """The single parent contract/customer for this child."""

    payer: Optional[Literal["SELF", "PARENT"]] = None
    """
    Account hierarchy M3 - Indicates which customer should pay for the child's
    invoice charges

    **SELF**: The child pays for its own invoice charges

    **PARENT**: The parent pays for the child's invoice charges
    """

    usage_statement_behavior: Optional[Literal["CONSOLIDATE", "SEPARATE"]] = None
    """
    Account hierarchy M3 - Indicates the behavior of the child's invoice statements
    on the parent's invoices.

    **CONSOLIDATE**: Child's invoice statements will be added to parent's
    consolidated invoices

    **SEPARATE**: Child's invoice statements will appear not appear on parent's
    consolidated invoices
    """


HierarchyConfiguration: TypeAlias = Union[ParentHierarchyConfiguration, ChildHierarchyConfiguration]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Optional
from typing_extensions import Literal

import httpx

from ...._types import Body, Omit, Query, Headers, NotGiven, omit, not_given
from ...._utils import maybe_transform, async_maybe_transform
from ...._compat import cached_property
from ...._resource import SyncAPIResource, AsyncAPIResource
from ...._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from ...._base_client import make_request_options
from ....types.v1.settings import billing_provider_list_params, billing_provider_create_params
from ....types.v1.settings.billing_provider_list_response import BillingProviderListResponse
from ....types.v1.settings.billing_provider_create_response import BillingProviderCreateResponse

__all__ = ["BillingProvidersResource", "AsyncBillingProvidersResource"]


class BillingProvidersResource(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> BillingProvidersResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/Metronome-Industries/metronome-python#accessing-raw-response-data-eg-headers
        """
        return BillingProvidersResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> BillingProvidersResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/Metronome-Industries/metronome-python#with_streaming_response
        """
        return BillingProvidersResourceWithStreamingResponse(self)

    def create(
        self,
        *,
        billing_provider: Literal["aws_marketplace", "azure_marketplace", "gcp_marketplace"],
        configuration: Dict[str, object],
        delivery_method: Literal["direct_to_billing_provider", "aws_sqs", "aws_sns"],
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> BillingProviderCreateResponse:
        """Set up account-level configuration for a billing provider.

        Once configured,
        individual contracts across customers can be mapped to this configuration using
        the returned delivery_method_id.

        Args:
          billing_provider: The billing provider set for this configuration.

          configuration: Account-level configuration for the billing provider. The structure of this
              object is specific to the billing provider and delivery provider combination.
              See examples below.

          delivery_method: The method to use for delivering invoices for this configuration.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._post(
            "/v1/setUpBillingProvider",
            body=maybe_transform(
                {
                    "billing_provider": billing_provider,
                    "configuration": configuration,
                    "delivery_method": delivery_method,
                },
                billing_provider_create_params.BillingProviderCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=BillingProviderCreateResponse,
        )

    def list(
        self,
        *,
        next_page: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> BillingProviderListResponse:
        """
        Lists all configured billing providers and their delivery method configurations
        for your account. Returns provider details, delivery method IDs, and
        configuration settings needed for mapping individual customer contracts to
        billing integrations.

        Args:
          next_page: The cursor to the next page of results

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._post(
            "/v1/listConfiguredBillingProviders",
            body=maybe_transform({"next_page": next_page}, billing_provider_list_params.BillingProviderListParams),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=BillingProviderListResponse,
        )


class AsyncBillingProvidersResource(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncBillingProvidersResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/Metronome-Industries/metronome-python#accessing-raw-response-data-eg-headers
        """
        return AsyncBillingProvidersResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncBillingProvidersResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/Metronome-Industries/metronome-python#with_streaming_response
        """
        return AsyncBillingProvidersResourceWithStreamingResponse(self)

    async def create(
        self,
        *,
        billing_provider: Literal["aws_marketplace", "azure_marketplace", "gcp_marketplace"],
        configuration: Dict[str, object],
        delivery_method: Literal["direct_to_billing_provider", "aws_sqs", "aws_sns"],
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> BillingProviderCreateResponse:
        """Set up account-level configuration for a billing provider.

        Once configured,
        individual contracts across customers can be mapped to this configuration using
        the returned delivery_method_id.

        Args:
          billing_provider: The billing provider set for this configuration.

          configuration: Account-level configuration for the billing provider. The structure of this
              object is specific to the billing provider and delivery provider combination.
              See examples below.

          delivery_method: The method to use for delivering invoices for this configuration.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._post(
            "/v1/setUpBillingProvider",
            body=await async_maybe_transform(
                {
                    "billing_provider": billing_provider,
                    "configuration": configuration,
                    "delivery_method": delivery_method,
                },
                billing_provider_create_params.BillingProviderCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=BillingProviderCreateResponse,
        )

    async def list(
        self,
        *,
        next_page: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> BillingProviderListResponse:
        """
        Lists all configured billing providers and their delivery method configurations
        for your account. Returns provider details, delivery method IDs, and
        configuration settings needed for mapping individual customer contracts to
        billing integrations.

        Args:
          next_page: The cursor to the next page of results

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._post(
            "/v1/listConfiguredBillingProviders",
            body=await async_maybe_transform(
                {"next_page": next_page}, billing_provider_list_params.BillingProviderListParams
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=BillingProviderListResponse,
        )


class BillingProvidersResourceWithRawResponse:
    def __init__(self, billing_providers: BillingProvidersResource) -> None:
        self._billing_providers = billing_providers

        self.create = to_raw_response_wrapper(
            billing_providers.create,
        )
        self.list = to_raw_response_wrapper(
            billing_providers.list,
        )


class AsyncBillingProvidersResourceWithRawResponse:
    def __init__(self, billing_providers: AsyncBillingProvidersResource) -> None:
        self._billing_providers = billing_providers

        self.create = async_to_raw_response_wrapper(
            billing_providers.create,
        )
        self.list = async_to_raw_response_wrapper(
            billing_providers.list,
        )


class BillingProvidersResourceWithStreamingResponse:
    def __init__(self, billing_providers: BillingProvidersResource) -> None:
        self._billing_providers = billing_providers

        self.create = to_streamed_response_wrapper(
            billing_providers.create,
        )
        self.list = to_streamed_response_wrapper(
            billing_providers.list,
        )


class AsyncBillingProvidersResourceWithStreamingResponse:
    def __init__(self, billing_providers: AsyncBillingProvidersResource) -> None:
        self._billing_providers = billing_providers

        self.create = async_to_streamed_response_wrapper(
            billing_providers.create,
        )
        self.list = async_to_streamed_response_wrapper(
            billing_providers.list,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, Optional
from datetime import datetime

from ..._models import BaseModel
from .schedule_point_in_time import SchedulePointInTime

__all__ = ["ScheduledCharge", "Product"]


class Product(BaseModel):
    id: str

    name: str


class ScheduledCharge(BaseModel):
    id: str

    product: Product

    schedule: SchedulePointInTime

    archived_at: Optional[datetime] = None

    custom_fields: Optional[Dict[str, str]] = None
    """Custom fields to be added eg. { "key1": "value1", "key2": "value2" }"""

    name: Optional[str] = None
    """displayed on invoices"""

    netsuite_sales_order_id: Optional[str] = None
    """This field's availability is dependent on your client's configuration."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import pytest

from metronome import Metronome, AsyncMetronome
from tests.utils import assert_matches_type
from metronome.types.v1 import (
    InvoiceVoidResponse,
    InvoiceRegenerateResponse,
)

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestInvoices:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @parametrize
    def test_method_regenerate(self, client: Metronome) -> None:
        invoice = client.v1.invoices.regenerate(
            id="6a37bb88-8538-48c5-b37b-a41c836328bd",
        )
        assert_matches_type(InvoiceRegenerateResponse, invoice, path=["response"])

    @parametrize
    def test_raw_response_regenerate(self, client: Metronome) -> None:
        response = client.v1.invoices.with_raw_response.regenerate(
            id="6a37bb88-8538-48c5-b37b-a41c836328bd",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(InvoiceRegenerateResponse, invoice, path=["response"])

    @parametrize
    def test_streaming_response_regenerate(self, client: Metronome) -> None:
        with client.v1.invoices.with_streaming_response.regenerate(
            id="6a37bb88-8538-48c5-b37b-a41c836328bd",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = response.parse()
            assert_matches_type(InvoiceRegenerateResponse, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_method_void(self, client: Metronome) -> None:
        invoice = client.v1.invoices.void(
            id="6a37bb88-8538-48c5-b37b-a41c836328bd",
        )
        assert_matches_type(InvoiceVoidResponse, invoice, path=["response"])

    @parametrize
    def test_raw_response_void(self, client: Metronome) -> None:
        response = client.v1.invoices.with_raw_response.void(
            id="6a37bb88-8538-48c5-b37b-a41c836328bd",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(InvoiceVoidResponse, invoice, path=["response"])

    @parametrize
    def test_streaming_response_void(self, client: Metronome) -> None:
        with client.v1.invoices.with_streaming_response.void(
            id="6a37bb88-8538-48c5-b37b-a41c836328bd",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = response.parse()
            assert_matches_type(InvoiceVoidResponse, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True


class TestAsyncInvoices:
    parametrize = pytest.mark.parametrize(
        "async_client", [False, True, {"http_client": "aiohttp"}], indirect=True, ids=["loose", "strict", "aiohttp"]
    )

    @parametrize
    async def test_method_regenerate(self, async_client: AsyncMetronome) -> None:
        invoice = await async_client.v1.invoices.regenerate(
            id="6a37bb88-8538-48c5-b37b-a41c836328bd",
        )
        assert_matches_type(InvoiceRegenerateResponse, invoice, path=["response"])

    @parametrize
    async def test_raw_response_regenerate(self, async_client: AsyncMetronome) -> None:
        response = await async_client.v1.invoices.with_raw_response.regenerate(
            id="6a37bb88-8538-48c5-b37b-a41c836328bd",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = await response.parse()
        assert_matches_type(InvoiceRegenerateResponse, invoice, path=["response"])

    @parametrize
    async def test_streaming_response_regenerate(self, async_client: AsyncMetronome) -> None:
        async with async_client.v1.invoices.with_streaming_response.regenerate(
            id="6a37bb88-8538-48c5-b37b-a41c836328bd",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = await response.parse()
            assert_matches_type(InvoiceRegenerateResponse, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_method_void(self, async_client: AsyncMetronome) -> None:
        invoice = await async_client.v1.invoices.void(
            id="6a37bb88-8538-48c5-b37b-a41c836328bd",
        )
        assert_matches_type(InvoiceVoidResponse, invoice, path=["response"])

    @parametrize
    async def test_raw_response_void(self, async_client: AsyncMetronome) -> None:
        response = await async_client.v1.invoices.with_raw_response.void(
            id="6a37bb88-8538-48c5-b37b-a41c836328bd",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = await response.parse()
        assert_matches_type(InvoiceVoidResponse, invoice, path=["response"])

    @parametrize
    async def test_streaming_response_void(self, async_client: AsyncMetronome) -> None:
        async with async_client.v1.invoices.with_streaming_response.void(
            id="6a37bb88-8538-48c5-b37b-a41c836328bd",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = await response.parse()
            assert_matches_type(InvoiceVoidResponse, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Union, Iterable
from datetime import datetime
from typing_extensions import Literal

import httpx

from ....._types import Body, Omit, Query, Headers, NotGiven, omit, not_given
from ....._utils import maybe_transform, async_maybe_transform
from ....._compat import cached_property
from ....._resource import SyncAPIResource, AsyncAPIResource
from ....._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from .....pagination import SyncCursorPage, AsyncCursorPage
from ....._base_client import AsyncPaginator, make_request_options
from .....types.shared_params.tier import Tier
from .....types.v1.contracts.rate_cards import rate_add_params, rate_list_params, rate_add_many_params
from .....types.shared_params.commit_rate import CommitRate
from .....types.v1.contracts.rate_cards.rate_add_response import RateAddResponse
from .....types.v1.contracts.rate_cards.rate_list_response import RateListResponse
from .....types.v1.contracts.rate_cards.rate_add_many_response import RateAddManyResponse

__all__ = ["RatesResource", "AsyncRatesResource"]


class RatesResource(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> RatesResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/Metronome-Industries/metronome-python#accessing-raw-response-data-eg-headers
        """
        return RatesResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> RatesResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/Metronome-Industries/metronome-python#with_streaming_response
        """
        return RatesResourceWithStreamingResponse(self)

    def list(
        self,
        *,
        at: Union[str, datetime],
        rate_card_id: str,
        limit: int | Omit = omit,
        next_page: str | Omit = omit,
        selectors: Iterable[rate_list_params.Selector] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SyncCursorPage[RateListResponse]:
        """
        Understand the rate schedule at a given timestamp, optionally filtering the list
        of rates returned based on properties such as `product_id` and
        `pricing_group_values`. For example, you may want to display the current price
        for a given product in your product experience - use this endpoint to fetch that
        information from its source of truth in Metronome.

        If you want to understand the rates for a specific customer's contract,
        inclusive of contract-level overrides, use the `getContractRateSchedule`
        endpoint.

        Args:
          at: inclusive starting point for the rates schedule

          rate_card_id: ID of the rate card to get the schedule for

          limit: Max number of results that should be returned

          next_page: Cursor that indicates where the next page of results should start.

          selectors: List of rate selectors, rates matching ANY of the selector will be included in
              the response Passing no selectors will result in all rates being returned.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get_api_list(
            "/v1/contract-pricing/rate-cards/getRates",
            page=SyncCursorPage[RateListResponse],
            body=maybe_transform(
                {
                    "at": at,
                    "rate_card_id": rate_card_id,
                    "selectors": selectors,
                },
                rate_list_params.RateListParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "limit": limit,
                        "next_page": next_page,
                    },
                    rate_list_params.RateListParams,
                ),
            ),
            model=RateListResponse,
            method="post",
        )

    def add(
        self,
        *,
        entitled: bool,
        product_id: str,
        rate_card_id: str,
        rate_type: Literal["FLAT", "PERCENTAGE", "SUBSCRIPTION", "TIERED", "CUSTOM"],
        starting_at: Union[str, datetime],
        billing_frequency: Literal["MONTHLY", "QUARTERLY", "ANNUAL", "WEEKLY"] | Omit = omit,
        commit_rate: CommitRate | Omit = omit,
        credit_type_id: str | Omit = omit,
        custom_rate: Dict[str, object] | Omit = omit,
        ending_before: Union[str, datetime] | Omit = omit,
        is_prorated: bool | Omit = omit,
        price: float | Omit = omit,
        pricing_group_values: Dict[str, str] | Omit = omit,
        quantity: float | Omit = omit,
        tiers: Iterable[Tier] | Omit = omit,
        use_list_prices: bool | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> RateAddResponse:
        """
        Add a new rate

        Args:
          product_id: ID of the product to add a rate for

          rate_card_id: ID of the rate card to update

          starting_at: inclusive effective date

          billing_frequency: Optional. Frequency to bill subscriptions with. Required for subscription type
              products with Flat rate.

          commit_rate: A distinct rate on the rate card. You can choose to use this rate rather than
              list rate when consuming a credit or commit.

          credit_type_id: The Metronome ID of the credit type to associate with price, defaults to USD
              (cents) if not passed. Used by all rate_types except type PERCENTAGE. PERCENTAGE
              rates use the credit type of associated rates.

          custom_rate: Only set for CUSTOM rate_type. This field is interpreted by custom rate
              processors.

          ending_before: exclusive end date

          is_prorated: Default proration configuration. Only valid for SUBSCRIPTION rate_type. Must be
              set to true.

          price: Default price. For FLAT and SUBSCRIPTION rate_type, this must be >=0. For
              PERCENTAGE rate_type, this is a decimal fraction, e.g. use 0.1 for 10%; this
              must be >=0 and <=1.

          pricing_group_values: Optional. List of pricing group key value pairs which will be used to calculate
              the price.

          quantity: Default quantity. For SUBSCRIPTION rate_type, this must be >=0.

          tiers: Only set for TIERED rate_type.

          use_list_prices: Only set for PERCENTAGE rate_type. Defaults to false. If true, rate is computed
              using list prices rather than the standard rates for this product on the
              contract.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._post(
            "/v1/contract-pricing/rate-cards/addRate",
            body=maybe_transform(
                {
                    "entitled": entitled,
                    "product_id": product_id,
                    "rate_card_id": rate_card_id,
                    "rate_type": rate_type,
                    "starting_at": starting_at,
                    "billing_frequency": billing_frequency,
                    "commit_rate": commit_rate,
                    "credit_type_id": credit_type_id,
                    "custom_rate": custom_rate,
                    "ending_before": ending_before,
                    "is_prorated": is_prorated,
                    "price": price,
                    "pricing_group_values": pricing_group_values,
                    "quantity": quantity,
                    "tiers": tiers,
                    "use_list_prices": use_list_prices,
                },
                rate_add_params.RateAddParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=RateAddResponse,
        )

    def add_many(
        self,
        *,
        rate_card_id: str,
        rates: Iterable[rate_add_many_params.Rate],
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> RateAddManyResponse:
        """
        Add new rates

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._post(
            "/v1/contract-pricing/rate-cards/addRates",
            body=maybe_transform(
                {
                    "rate_card_id": rate_card_id,
                    "rates": rates,
                },
                rate_add_many_params.RateAddManyParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=RateAddManyResponse,
        )


class AsyncRatesResource(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncRatesResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/Metronome-Industries/metronome-python#accessing-raw-response-data-eg-headers
        """
        return AsyncRatesResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncRatesResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/Metronome-Industries/metronome-python#with_streaming_response
        """
        return AsyncRatesResourceWithStreamingResponse(self)

    def list(
        self,
        *,
        at: Union[str, datetime],
        rate_card_id: str,
        limit: int | Omit = omit,
        next_page: str | Omit = omit,
        selectors: Iterable[rate_list_params.Selector] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> AsyncPaginator[RateListResponse, AsyncCursorPage[RateListResponse]]:
        """
        Understand the rate schedule at a given timestamp, optionally filtering the list
        of rates returned based on properties such as `product_id` and
        `pricing_group_values`. For example, you may want to display the current price
        for a given product in your product experience - use this endpoint to fetch that
        information from its source of truth in Metronome.

        If you want to understand the rates for a specific customer's contract,
        inclusive of contract-level overrides, use the `getContractRateSchedule`
        endpoint.

        Args:
          at: inclusive starting point for the rates schedule

          rate_card_id: ID of the rate card to get the schedule for

          limit: Max number of results that should be returned

          next_page: Cursor that indicates where the next page of results should start.

          selectors: List of rate selectors, rates matching ANY of the selector will be included in
              the response Passing no selectors will result in all rates being returned.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get_api_list(
            "/v1/contract-pricing/rate-cards/getRates",
            page=AsyncCursorPage[RateListResponse],
            body=maybe_transform(
                {
                    "at": at,
                    "rate_card_id": rate_card_id,
                    "selectors": selectors,
                },
                rate_list_params.RateListParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "limit": limit,
                        "next_page": next_page,
                    },
                    rate_list_params.RateListParams,
                ),
            ),
            model=RateListResponse,
            method="post",
        )

    async def add(
        self,
        *,
        entitled: bool,
        product_id: str,
        rate_card_id: str,
        rate_type: Literal["FLAT", "PERCENTAGE", "SUBSCRIPTION", "TIERED", "CUSTOM"],
        starting_at: Union[str, datetime],
        billing_frequency: Literal["MONTHLY", "QUARTERLY", "ANNUAL", "WEEKLY"] | Omit = omit,
        commit_rate: CommitRate | Omit = omit,
        credit_type_id: str | Omit = omit,
        custom_rate: Dict[str, object] | Omit = omit,
        ending_before: Union[str, datetime] | Omit = omit,
        is_prorated: bool | Omit = omit,
        price: float | Omit = omit,
        pricing_group_values: Dict[str, str] | Omit = omit,
        quantity: float | Omit = omit,
        tiers: Iterable[Tier] | Omit = omit,
        use_list_prices: bool | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> RateAddResponse:
        """
        Add a new rate

        Args:
          product_id: ID of the product to add a rate for

          rate_card_id: ID of the rate card to update

          starting_at: inclusive effective date

          billing_frequency: Optional. Frequency to bill subscriptions with. Required for subscription type
              products with Flat rate.

          commit_rate: A distinct rate on the rate card. You can choose to use this rate rather than
              list rate when consuming a credit or commit.

          credit_type_id: The Metronome ID of the credit type to associate with price, defaults to USD
              (cents) if not passed. Used by all rate_types except type PERCENTAGE. PERCENTAGE
              rates use the credit type of associated rates.

          custom_rate: Only set for CUSTOM rate_type. This field is interpreted by custom rate
              processors.

          ending_before: exclusive end date

          is_prorated: Default proration configuration. Only valid for SUBSCRIPTION rate_type. Must be
              set to true.

          price: Default price. For FLAT and SUBSCRIPTION rate_type, this must be >=0. For
              PERCENTAGE rate_type, this is a decimal fraction, e.g. use 0.1 for 10%; this
              must be >=0 and <=1.

          pricing_group_values: Optional. List of pricing group key value pairs which will be used to calculate
              the price.

          quantity: Default quantity. For SUBSCRIPTION rate_type, this must be >=0.

          tiers: Only set for TIERED rate_type.

          use_list_prices: Only set for PERCENTAGE rate_type. Defaults to false. If true, rate is computed
              using list prices rather than the standard rates for this product on the
              contract.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._post(
            "/v1/contract-pricing/rate-cards/addRate",
            body=await async_maybe_transform(
                {
                    "entitled": entitled,
                    "product_id": product_id,
                    "rate_card_id": rate_card_id,
                    "rate_type": rate_type,
                    "starting_at": starting_at,
                    "billing_frequency": billing_frequency,
                    "commit_rate": commit_rate,
                    "credit_type_id": credit_type_id,
                    "custom_rate": custom_rate,
                    "ending_before": ending_before,
                    "is_prorated": is_prorated,
                    "price": price,
                    "pricing_group_values": pricing_group_values,
                    "quantity": quantity,
                    "tiers": tiers,
                    "use_list_prices": use_list_prices,
                },
                rate_add_params.RateAddParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=RateAddResponse,
        )

    async def add_many(
        self,
        *,
        rate_card_id: str,
        rates: Iterable[rate_add_many_params.Rate],
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> RateAddManyResponse:
        """
        Add new rates

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._post(
            "/v1/contract-pricing/rate-cards/addRates",
            body=await async_maybe_transform(
                {
                    "rate_card_id": rate_card_id,
                    "rates": rates,
                },
                rate_add_many_params.RateAddManyParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=RateAddManyResponse,
        )


class RatesResourceWithRawResponse:
    def __init__(self, rates: RatesResource) -> None:
        self._rates = rates

        self.list = to_raw_response_wrapper(
            rates.list,
        )
        self.add = to_raw_response_wrapper(
            rates.add,
        )
        self.add_many = to_raw_response_wrapper(
            rates.add_many,
        )


class AsyncRatesResourceWithRawResponse:
    def __init__(self, rates: AsyncRatesResource) -> None:
        self._rates = rates

        self.list = async_to_raw_response_wrapper(
            rates.list,
        )
        self.add = async_to_raw_response_wrapper(
            rates.add,
        )
        self.add_many = async_to_raw_response_wrapper(
            rates.add_many,
        )


class RatesResourceWithStreamingResponse:
    def __init__(self, rates: RatesResource) -> None:
        self._rates = rates

        self.list = to_streamed_response_wrapper(
            rates.list,
        )
        self.add = to_streamed_response_wrapper(
            rates.add,
        )
        self.add_many = to_streamed_response_wrapper(
            rates.add_many,
        )


class AsyncRatesResourceWithStreamingResponse:
    def __init__(self, rates: AsyncRatesResource) -> None:
        self._rates = rates

        self.list = async_to_streamed_response_wrapper(
            rates.list,
        )
        self.add = async_to_streamed_response_wrapper(
            rates.add,
        )
        self.add_many = async_to_streamed_response_wrapper(
            rates.add_many,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal

from ..._models import BaseModel

__all__ = ["RecurringCommitSubscriptionConfig", "ApplySeatIncreaseConfig"]


class ApplySeatIncreaseConfig(BaseModel):
    is_prorated: bool
    """Indicates whether a mid-period seat increase should be prorated."""


class RecurringCommitSubscriptionConfig(BaseModel):
    allocation: Literal["INDIVIDUAL", "POOLED"]

    apply_seat_increase_config: ApplySeatIncreaseConfig

    subscription_id: str



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union
from datetime import datetime
from typing_extensions import Required, Annotated, TypedDict

from ...._utils import PropertyInfo

__all__ = ["PlanEndParams"]


class PlanEndParams(TypedDict, total=False):
    customer_id: Required[str]

    customer_plan_id: Required[str]

    ending_before: Annotated[Union[str, datetime], PropertyInfo(format="iso8601")]
    """RFC 3339 timestamp for when the plan ends (exclusive) for this customer.

    Must be at 0:00 UTC (midnight). If not provided, the plan end date will be
    cleared.
    """

    void_invoices: bool
    """If true, plan end date can be before the last finalized invoice date.

    Any invoices generated after the plan end date will be voided.
    """

    void_stripe_invoices: bool
    """Only applicable when void_invoices is set to true.

    If true, for every invoice that is voided we will also attempt to void/delete
    the stripe invoice (if any). Stripe invoices will be voided if finalized or
    deleted if still in draft state.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .rates import (
    RatesResource,
    AsyncRatesResource,
    RatesResourceWithRawResponse,
    AsyncRatesResourceWithRawResponse,
    RatesResourceWithStreamingResponse,
    AsyncRatesResourceWithStreamingResponse,
)
from .rate_cards import (
    RateCardsResource,
    AsyncRateCardsResource,
    RateCardsResourceWithRawResponse,
    AsyncRateCardsResourceWithRawResponse,
    RateCardsResourceWithStreamingResponse,
    AsyncRateCardsResourceWithStreamingResponse,
)
from .product_orders import (
    ProductOrdersResource,
    AsyncProductOrdersResource,
    ProductOrdersResourceWithRawResponse,
    AsyncProductOrdersResourceWithRawResponse,
    ProductOrdersResourceWithStreamingResponse,
    AsyncProductOrdersResourceWithStreamingResponse,
)
from .named_schedules import (
    NamedSchedulesResource,
    AsyncNamedSchedulesResource,
    NamedSchedulesResourceWithRawResponse,
    AsyncNamedSchedulesResourceWithRawResponse,
    NamedSchedulesResourceWithStreamingResponse,
    AsyncNamedSchedulesResourceWithStreamingResponse,
)

__all__ = [
    "ProductOrdersResource",
    "AsyncProductOrdersResource",
    "ProductOrdersResourceWithRawResponse",
    "AsyncProductOrdersResourceWithRawResponse",
    "ProductOrdersResourceWithStreamingResponse",
    "AsyncProductOrdersResourceWithStreamingResponse",
    "RatesResource",
    "AsyncRatesResource",
    "RatesResourceWithRawResponse",
    "AsyncRatesResourceWithRawResponse",
    "RatesResourceWithStreamingResponse",
    "AsyncRatesResourceWithStreamingResponse",
    "NamedSchedulesResource",
    "AsyncNamedSchedulesResource",
    "NamedSchedulesResourceWithRawResponse",
    "AsyncNamedSchedulesResourceWithRawResponse",
    "NamedSchedulesResourceWithStreamingResponse",
    "AsyncNamedSchedulesResourceWithStreamingResponse",
    "RateCardsResource",
    "AsyncRateCardsResource",
    "RateCardsResourceWithRawResponse",
    "AsyncRateCardsResourceWithRawResponse",
    "RateCardsResourceWithStreamingResponse",
    "AsyncRateCardsResourceWithStreamingResponse",
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union
from datetime import datetime
from typing_extensions import Required, Annotated, TypedDict

from ..._utils import PropertyInfo

__all__ = ["CustomerListCostsParams"]


class CustomerListCostsParams(TypedDict, total=False):
    customer_id: Required[str]

    ending_before: Required[Annotated[Union[str, datetime], PropertyInfo(format="iso8601")]]
    """RFC 3339 timestamp (exclusive)"""

    starting_on: Required[Annotated[Union[str, datetime], PropertyInfo(format="iso8601")]]
    """RFC 3339 timestamp (inclusive)"""

    limit: int
    """Max number of results that should be returned"""

    next_page: str
    """Cursor that indicates where the next page of results should start."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List
from datetime import datetime

from ..._models import BaseModel
from .credit_ledger_entry import CreditLedgerEntry
from ..shared.credit_type_data import CreditTypeData

__all__ = ["CreditGrantListEntriesResponse", "Ledger", "LedgerEndingBalance", "LedgerStartingBalance"]


class LedgerEndingBalance(BaseModel):
    """the effective balances at the end of the specified time window"""

    effective_at: datetime
    """
    the ending_before request parameter (if supplied) or the current billing
    period's end date
    """

    excluding_pending: float
    """
    the ending balance, including the balance of all grants that have not expired
    before the effective_at date and deductions that happened before the
    effective_at date
    """

    including_pending: float
    """
    the excluding_pending balance plus any pending invoice deductions and
    expirations that will happen by the effective_at date
    """


class LedgerStartingBalance(BaseModel):
    effective_at: datetime
    """
    the starting_on request parameter (if supplied) or the first credit grant's
    effective_at date
    """

    excluding_pending: float
    """
    the starting balance, including all posted grants, deductions, and expirations
    that happened at or before the effective_at timestamp
    """

    including_pending: float
    """
    the excluding_pending balance plus any pending activity that has not been posted
    at the time of the query
    """


class Ledger(BaseModel):
    credit_type: CreditTypeData

    ending_balance: LedgerEndingBalance
    """the effective balances at the end of the specified time window"""

    entries: List[CreditLedgerEntry]

    pending_entries: List[CreditLedgerEntry]

    starting_balance: LedgerStartingBalance


class CreditGrantListEntriesResponse(BaseModel):
    customer_id: str

    ledgers: List[Ledger]
